import { DependencyManager } from "./dependency"
import "source-map-support/register"
import * as async from "async"
import * as fs from "fs"
import * as pathModule from "path"
import * as colors from "colors/safe"
import * as Leaf from "leaf-ts"
import { SharedCallbacks } from "./sharedCallbacks"
let useColors = colors


export type ServiceMatcher = string | RegExp
export class ServiceContext<TServices  = {}> {
    private BuiltInServiceInstances = {}
    public passiveServices = {}
    constructor(private settings = {}) {
        const BuiltInService = require("./builtInService")
        //for (let name in BuiltInService) {
        //    let Con = BuiltInService[name] as any
        //    let s = new Con(this.settings)
        //    this.BuiltInServiceInstances[s.name] = s
        //}
        // When dependency not met but we have that dependency in our built-in service
        this.dependencyManager.onAddDymanicDependency = (name) => {
            if (this.passiveServices[name]) {
                this.addToDependency(this.passiveServices[name])
                return
            }
            let Cons = BuiltInService[name]
            if (Cons && !this.BuiltInServiceInstances[name]) {
                this.BuiltInServiceInstances[name] = new Cons(settings)
            }
            let s = this.BuiltInServiceInstances[name]
            if (s) {
                if (s["asBuiltIn"]) {
                    s["asBuiltIn"]()
                }
                this.addToDependency(s)
            }
        }
    }
    services: TServices = {} as any
    dependencyManager: DependencyManager = new DependencyManager;
    register(s: Service): void {
        console.log("Register service", useColors.green(s.name))
        // passive module will only add when required
        if (s.passive) {
            this.passiveServices[s.name] = s
        } else {
            this.addToDependency(s)
        }
    }
    addToDependency(s: Service) {
        this.dependencyManager.add(s.name, s.dependencies || null)
        this.services[s.name] = s
    }
    supportedLoadExtnames = [".ts", ".js"]
    loadServices(path: string) {
        path = pathModule.resolve(path)
        let files = fs.readdirSync(path)
        files = files.filter((item) => {
            let ext = pathModule.extname(item)
            if (this.supportedLoadExtnames.indexOf(ext) < 0) {
                return false
            }
            return true
        })
        files = files.map((item) => {
            let ext = pathModule.extname(item)
            let base = pathModule.basename(item, ext)
            return pathModule.join(path, base)
        })
        let toLoads = []
        for (let file of files) {
            if (toLoads.indexOf(file) >= 0) continue
            toLoads.push(file)
        }
        for (let file of toLoads) {
            this.loadService(file)
        }
    }
    loadService(path: string) {
        let result = require(path) || {}
        for (let name in result) {
            let obj = result[name]
            if (this.isServiceConstructor(obj)) {
                let Cons = obj as any
                let s = new Cons(this.settings)
                if (!s.name) {
                    console.log(useColors.cyan(`Detect service without name. Use constructor name:${Cons.name}`))
                    s.name = Cons.name
                }
                console.log(`Load service ${useColors.green(s.name)}`)
                this.register(s)
            }
        }
    }
    private isServiceConstructor(obj) {
        return typeof obj === "function" && obj.prototype["thisObjectIsRootServiceConstructor"]
    }
    public maxInitializingTimeout = 1000 * 60
    isSetup: boolean = false
    private finishSetup = SharedCallbacks.create()
    setup(finish: (err?: Error) => void = () => null): void {
        this.finishSetup.push(finish)
        if (this.isSetup) {
            return
        }
        this.isSetup = true
        let initialSequence = this.dependencyManager.resolve()
        async.eachSeries(initialSequence, (name, _done) => {
            let service = this.services[name]
            let done = (err?: Error) => {
                if (err) {
                    _done(err)
                    return
                }
                this.serviceAfterInitialized(service, _done)
            }
            if (!service) {
                done(new Error("Fail to find service name " + name))
                return
            }
            if (service.isInitialized) {
                done()
                return
            }
            service.services = this.services
            console.log(`Initializing ${useColors.green(name)}`)
            if (service.initialize) {
                this.initializeService(service, done)
            } else {
                // this service don't need initialization
                service.isInitialized = true
                this.logInitialized(name)
                done()
            }
        }, (err: Error) => {
            this.isSetup = false
            if (err) {
                this.finishSetup(err)
                return
            }
            console.log(useColors.bgGreen("All service initialized"))
            this.finishSetup()
        })
    }
    private serviceAfterInitialized(service: Service, done: Function) {
        let timer = null
        let name = service.name
        if (service.afterInitialize.length == 1) {
            timer = setTimeout(() => {
                console.error(`Service ${name} after initialize hook timeout, exit program`)
                process.exit(1)
            }, this.maxInitializingTimeout)
            service.afterInitialize((err?: Error) => {
                clearTimeout(timer)
                if (err) {
                    console.error(useColors.red(`Fail to afterInitialize service ${name}.`))
                    done(err)
                    return
                }
                done()
            })
        } else {
            service.afterInitialize()
            done()
        }
    }
    private initializeService(service: Service, done: Function) {
        let name = service.name
        let timer = null
        if (service.initialize.length == 1) {
            timer = setTimeout(() => {
                console.error(`Service ${name} initializing timeout, exit program`)
                process.exit(1)
            }, this.maxInitializingTimeout)
            service.initialize((err?: Error) => {
                clearTimeout(timer)
                if (err) {
                    console.error(useColors.red(`Fail to initialize service ${name}.`))
                    done(err)
                    return
                }
                service.isInitialized = true
                this.logInitialized(name)
                done()
            })
        } else {
            service.initialize()
            service.isInitialized = true
            this.logInitialized(name)
            done()
        }
    }
    private logInitialized(name: string) {
        console.log(`Service ${useColors.green(name)} Initialized`)
    }
}

export abstract class Service<TSettings = any> {
    abstract name: string
    abstract services
    passive = false
    dependencies: ServiceMatcher[] = []
    befores: ServiceMatcher[] = []
    isInitialized: boolean = false
    initialize(done?: (err?: Error) => void) {
        done()
    }
    afterInitialize(done?: (err?: Error) => void) {
        done()
    }
    constructor(public readonly _defaultSettings: TSettings, settings?: TSettings) {
        this.buildOption(_defaultSettings, settings)
    }
    settings: TSettings = {} as TSettings
    protected buildOption(df, config) {
        this.settings = Leaf.Util.buildOption(df, config)
    }
}

Service.prototype["thisObjectIsRootServiceConstructor"] = true
