import { Service } from "../lib/service"
import { MemoryLockService } from "./memoryLockService"
import { MongodbService } from "./mongodbService"
import * as mongodb from "mongodb"
import * as Leaf from "leaf-ts"

export type Callback<T = any, E extends Error = Error> = (e?: E, result?: T) => void
export const Errors = Leaf.ErrorDoc.build({
    UnknownError: {},
    NotFound: {},
    InvalidParameter: {},
})

export class IncrementalIdService extends Service {
    name = "IncrementalIdService"
    dependencies = ["MongodbService"]
    services: {
        MongodbService: MongodbService
    }
    constructor() {
        super({}, {})
    }
    protected collectionName = "IncrementalIdService"
    protected incrementalIdCollection: mongodb.Collection
    public readonly defaultOffset = 1
    initialize() {
        this.incrementalIdCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.incrementalIdCollection.createIndex({
            name: 1
        }, {
                unique: true
            })
    }
    availableItems = {}
    ensure(option: { name: string, offset: number }, callback: Callback<void>) {
        if (!option.name) {
            callback(new Errors.InvalidParameter)
            return
        }
        this.incrementalIdCollection.insertOne({
            name: option.name,
            offset: option.offset | this.defaultOffset
        }, (err) => {
            if (err && err["code"] !== 11000) {
                callback(new Errors.UnknownError("Unknown error when insert base document", { via: err }))
                return
            }
            this.availableItems[option.name] = true
            callback(null)
        })
    }
    offset(option: { name: string, offset: number }, callback: Callback<void>) {
        if (!option.name) {
            callback(new Errors.InvalidParameter)
            return
        }
        this.incrementalIdCollection.findOneAndUpdate({
            name: option.name,
        }, {
                $set: {
                    offset: option.offset
                }
            }, (err) => {
                callback(null)
            })
    }
    next(option: { name: string, count?: number }, callback: Callback<number>) {
        if (!option.name) {
            callback(new Errors.InvalidParameter)
            return
        }
        if (!this.availableItems[option.name]) {
            callback(new Errors.NotFound())
            return
        }
        let count = option.count || 1
        if (count < 0) count = 1
        this.incrementalIdCollection.findOneAndUpdate({
            name: option.name
        }, {
                $inc: {
                    offset: count
                }
            }, (err, result) => {
                if (err) {
                    callback(err)
                    return
                }
                callback(null, result.value.offset)
            })
    }
}
