import { Service } from "../lib/service"
import * as Leaf from "leaf-ts"

export const Errors = Leaf.ErrorDoc.create()
    .define("InvalidLegacyUsername")
    .define("LegacyUsernameTooShort")
    .define("LegacyUsernameTooLong")
    .define("InvalidLegacyUsername")
    .define("PasswordTooShort")
    .define("PasswordTooLong")
    .define("InvalidPassword")
    .define("InvalidEmail")
    .define("EmailTooLong")
    .define("InvalidDisplayName")
    .define("DisplayNameTooShort")
    .define("DisplayNameTooLong")
    .define("LegacyUsernameRequired")
    .define("PasswordRequired")
    .define("EmailRequired")
    .generate() as Leaf.ErrorsOf<{
        InvalidLegacyUsername: {}
        LegacyUsernameTooShort: {}
        LegacyUsernameTooLong: {}
        PasswordTooShort: {}
        PasswordTooLong: {}
        EmailTooLong: {}
        InvalidEmail: {}
        InvalidDisplayName: {}
        DisplayNameTooShort: {}
        DisplayNameTooLong: {}
        InvalidPassword: {}
        LegacyUsernameRequired: {}
        PasswordRequired: {}
        EmailRequired: {}
    }>
export type LegacyUser = LegacyUserDefinitionService.LegacyUser
export class LegacyUserDefinitionService<TLegacyUser extends LegacyUserDefinitionService.LegacyUser> extends Service {
    readonly name = "LegacyUserDefinitionService"
    services: null
    public LegacyUserPartialValidator: Leaf.Validator = GenerateLegacyUserValidator()
    public LegacyUserValidator = this.LegacyUserPartialValidator.clone()
        .field("legacyUsername").exists().error(new Errors.LegacyUsernameRequired)
        .field("email").exists().error(new Errors.EmailRequired)
        .field("password").exists().error(new Errors.PasswordRequired)
}
export namespace LegacyUserDefinitionService {
    export interface LegacyUser {
        id?: string
        legacyUsername: string
        displayName?: string
        passwordHash?: string
        password?: string
        email: string
        createAt?: number
    }
}

export interface LegacyUserValidationConfig {
    minLegacyUsernameLength: number
    maxLegacyUsernameLength: number
    minDisplayNameLength: number
    maxDisplayNameLength: number
    minPasswordLength: number
    maxPasswordLength: number
    maxEmailLength: number
}
export const DefaultLegacyUserValidationConfig: LegacyUserValidationConfig = {
    minLegacyUsernameLength: 3,
    maxLegacyUsernameLength: 32,
    minDisplayNameLength: 3,
    maxDisplayNameLength: 32,
    minPasswordLength: 8,
    maxPasswordLength: 256,
    maxEmailLength: 128
}
export function GenerateLegacyUserValidator(config: LegacyUserValidationConfig = DefaultLegacyUserValidationConfig) {
    return Leaf.Validator.create()
        .field("legacyUsername")
        .string().error(new Errors.InvalidLegacyUsername)
        .gt(config.minLegacyUsernameLength - 1).error(new Errors.LegacyUsernameTooShort)
        .lt(config.maxLegacyUsernameLength + 1).error(new Errors.LegacyUsernameTooLong)
        .match(/^[0-9a-z_]+$/i).error(new Errors.InvalidLegacyUsername)

        .field("displayName")
        .string().error(new Errors.InvalidDisplayName)
        .gt(config.minDisplayNameLength - 1).error(new Errors.DisplayNameTooShort)
        .lt(config.maxDisplayNameLength + 1).error(new Errors.DisplayNameTooLong)

        .field("password")
        .string().error(new Errors.InvalidPassword)
        .gt(config.minPasswordLength - 1).error(new Errors.PasswordTooShort)
        .lt(config.maxPasswordLength + 1).error(new Errors.PasswordTooLong)

        .field("email")
        .lt(config.maxEmailLength + 1).error(new Errors.EmailTooLong)
        .match(/^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/)
        .readonly()
}
