import { Service } from "../lib/service"
import { LegacyUserDefinitionService, LegacyUser } from "./legacyUserDefinitionService"
import { MongodbService } from "./mongodbService"
import * as mongodb from "mongodb"
import * as crypto from "crypto"
import * as Leaf from "leaf-ts"
export type Callback<T> = (err?: Error, result?: T) => void

export const Errors = Leaf.ErrorDoc.create()
    .define("UnknownError")
    .define("DBError")
    .define("AlreadyExists")
    .define("LegacyUsernameAlreadyExists")
    .define("EmailAlreadyExists")
    .define("InvalidParameter")
    .generate() as Leaf.ErrorsOf<{
        UnknownError: {}
        DBError: {}
        AlreadyExists: {}
        InvalidParameter: {}
        LegacyUsernameAlreadyExists: {}
        EmailAlreadyExists: {}
    }>
function clearLegacyUserPassword(obj: any, name: string, descriptor: TypedPropertyDescriptor<Function>) {
    let fn = descriptor.value
    descriptor.value = function (option, callback: Function) {
        fn.call(this, option, (err, value) => {
            if (value instanceof Array) {
                value = value.forEach((item) => {
                    delete item.passwordHash
                })
            } else if (value) {
                delete value.passwordHash
            }
            callback(err, value)
        })
    }
}
export class LegacyUserService<TLegacyUser extends LegacyUser = LegacyUser> extends Service {
    readonly name = "LegacyUserService"
    readonly dependencies = ["MongodbService", "LegacyUserDefinitionService"] as any
    readonly collectionName = "legacyUser"
    public readonly Errors = Errors
    constructor(config: {
        legacyUser?: {
            salt?: string
            blacklist?: string[]
        }
    } = {}) {
        super({}, {})
        if (config.legacyUser) {
            if (config.legacyUser.salt) {
                this.salt = config.legacyUser.salt
            }
            if (config.legacyUser.blacklist) {
                this.blacklist = config.legacyUser.blacklist
            }
        }
    }
    private blacklist: string[] = []
    private salt: string = "default-legacyUser-secret-salt"
    services: {
        MongodbService: MongodbService
        LegacyUserDefinitionService: LegacyUserDefinitionService<TLegacyUser>
    }
    legacyUserCollection: mongodb.Collection
    initialize() {
        this.salt
        this.legacyUserCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.legacyUserCollection.createIndex({
            legacyUsername: 1
        }, {
                unique: true
            })
        this.legacyUserCollection.createIndex({
            email: 1
        }, {
                unique: true
            })
        this.legacyUserCollection.createIndex({
            passwordHash: 1
        })
    }
    private createPasswordHashString(rawPassword: string): string {
        let level1 = crypto.createHash("sha256").update(rawPassword + this.salt).digest('hex')
        let level2 = crypto.createHash("sha512").update(level1 + this.salt).digest('hex')
        return level2
    }
    private translateMongoId(doc: { _id: mongodb.ObjectID } & TLegacyUser): TLegacyUser {
        if (!doc) return null
        let id = doc._id
        delete doc._id
        doc.id = id.toString()
        return doc
    }
    public toPublicLegacyUser(legacyUser: TLegacyUser): TLegacyUser {
        delete legacyUser.passwordHash
        delete legacyUser.password
        return legacyUser
    }
    @clearLegacyUserPassword
    getLegacyUserById(id: string, callback: Callback<TLegacyUser>) {
        let _id = new mongodb.ObjectID(id)
        this.legacyUserCollection.findOne({ _id }, (error: Error, doc: { _id: mongodb.ObjectID } & TLegacyUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let legacyUser = this.translateMongoId(doc)
            callback(null, legacyUser)
        })
    }
    @clearLegacyUserPassword
    getLegacyUserByLegacyUsername(legacyUsername: string, callback: Callback<TLegacyUser>) {
        legacyUsername = legacyUsername.toLowerCase()
        this.legacyUserCollection.findOne({ legacyUsername }, (error: Error, doc: { _id: mongodb.ObjectID } & TLegacyUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let legacyUser = this.translateMongoId(doc)
            callback(null, legacyUser)
        })
    }
    @clearLegacyUserPassword
    getLegacyUserByEmail(email: string, callback: Callback<TLegacyUser>) {
        email = email.toLowerCase()
        this.legacyUserCollection.findOne({ email }, (error: Error, doc: { _id: mongodb.ObjectID } & TLegacyUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let legacyUser = this.translateMongoId(doc)
            callback(null, legacyUser)
        })
    }
    @clearLegacyUserPassword
    getLegacyUserByLogin(option: { login: string, password: string }, callback: Callback<TLegacyUser>) {
        let login = option.login.toLowerCase()
        let passwordHash = this.createPasswordHashString(option.password)
        this.legacyUserCollection.findOne({ passwordHash: passwordHash, $or: [{ legacyUsername: login }, { email: login }] }, (error: Error, doc: { _id: mongodb.ObjectID } & TLegacyUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let legacyUser = this.translateMongoId(doc)
            callback(null, legacyUser)
        })
    }

    @clearLegacyUserPassword
    createLegacyUser(partialLegacyUser: Partial<TLegacyUser>, callback: Callback<TLegacyUser>) {
        if (this.blacklist.indexOf(partialLegacyUser.legacyUsername) >= 0) {
            callback(new Errors.InvalidParameter("Invalid legacyUsername"))
            return
        }
        if (!partialLegacyUser.legacyUsername || !partialLegacyUser.email) {
            callback(new Errors.InvalidParameter("Empty legacyUsername or email"))
            return
        }
        // use legacyUsername as display name by default
        if (!partialLegacyUser.displayName) {
            partialLegacyUser.displayName = partialLegacyUser.legacyUsername
        }
        // then convert legacyUsername to all lowercase
        partialLegacyUser.legacyUsername = partialLegacyUser.legacyUsername.toLocaleLowerCase()
        partialLegacyUser.email = partialLegacyUser.email.toLowerCase()
        try {
        }
        catch (error) {
            callback(error)
            return
        }
        try {
            this.services.LegacyUserDefinitionService.LegacyUserValidator.check(partialLegacyUser)
        } catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        let passwordHash = this.createPasswordHashString(partialLegacyUser.password)
        let legacyUser = {
            legacyUsername: partialLegacyUser.legacyUsername,
            displayName: partialLegacyUser.displayName,
            passwordHash: passwordHash,
            email: partialLegacyUser.email,
            crteateAt: Date.now()
        }
        this.getLegacyUserByLegacyUsername(legacyUser.legacyUsername, (error: Error, doc: TLegacyUser) => {
            if (doc) {
                callback(new Errors.LegacyUsernameAlreadyExists())
                return
            }
            this.getLegacyUserByEmail(legacyUser.email, (error: Error, doc: TLegacyUser) => {
                if (doc) {
                    callback(new Errors.EmailAlreadyExists())
                    return
                }
                this.legacyUserCollection.insert(legacyUser, (error: Error, doc: mongodb.InsertOneWriteOpResult) => {
                    if (error) {
                        // on a edge case that 2 legacyUser try to use a same legacyUsername and something like that
                        callback(new Errors.DBError("DatabaseError", { via: error }))
                        return
                    }
                    let legacyUser = this.translateMongoId(doc.ops[0])
                    callback(null, legacyUser)
                })
            })
        })
    }
    @clearLegacyUserPassword
    updateLegacyUserPasswordByLegacyUsername(option: { legacyUsername: string, password: string }, callback: Callback<TLegacyUser>) {
        try {
            this.services.LegacyUserDefinitionService.LegacyUserValidator.check({ password: option.password })
        }
        catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        const legacyUsername = option.legacyUsername.toLowerCase()
        let passwordHash = this.createPasswordHashString(option.password)
        this.legacyUserCollection.findOneAndUpdate({ legacyUsername }, { $set: { passwordHash } }, { returnOriginal: false }, (error, doc: mongodb.FindAndModifyWriteOpResultObject) => {
            if (error) {
                callback(error)
                return
            }
            let legacyUser = this.translateMongoId(doc.value)
            callback(null, legacyUser)
        })
    }
    @clearLegacyUserPassword
    updateLegacyUserPasswordById(option: { id: string, password: string }, callback: Callback<TLegacyUser>) {
        let _id = new mongodb.ObjectID(option.id)
        try {
            this.services.LegacyUserDefinitionService.LegacyUserValidator.check({ password: option.password })
        }
        catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        let passwordHash = this.createPasswordHashString(option.password)
        this.legacyUserCollection.findOneAndUpdate({ _id }, { $set: { passwordHash } }, { returnOriginal: false }, (error, doc: mongodb.FindAndModifyWriteOpResultObject) => {
            if (error) {
                callback(error)
                return
            }
            let legacyUser = this.translateMongoId(doc.value)
            callback(null, legacyUser)
        })
    }
}
