import * as Leaf from "leaf-ts"
import { Service } from "../lib/service"
import * as fsExtra from "fs-extra"
import * as multer from "multer"
import * as fs from "fs"
import * as express from "express"
import * as pathModule from "path"
import { BlobService, Blob } from "./blobService";

export const Errors = Leaf.ErrorDoc.build({
    UploadError: {},
})
//uploadBlobService
export namespace UploadService {
    export type Settings = {
        upload: {
            temp: string
            // let empty to support all extnames
            acceptedExtnames?: string[]
            maxSizeInMB?: number
            route?: string
        }
    }
}

export class UploadService extends Service<UploadService.Settings> {
    name = "UploadService"
    dependencies = ["ExpressService", "BlobService"]
    services: { ExpressService, BlobService }
    constructor(settings: UploadService.Settings) {
        super({
            upload: {
                temp: "/tmp/upload",
                route: "/upload",
                // default support all extnames
                acceptedExtnames: null,
                maxSizeInMB: 8
            }
        }, settings)
    }
    error(res: express.Response, err) {
        res.json({
            state: false,
            error: err
        })
    }
    success(res: express.Response, data) {
        res.json({
            state: true,
            data
        })
    }
    initialize() {
        let storageConfig = this.createMulterConfig()
        let multerMiddleware = multer({
            storage: storageConfig,
            limits: {
                fieldSize: this.settings.upload.maxSizeInMB
            }
        })
        this.services.ExpressService.server.post(this.settings.upload.route, this.createUploadHandler(multerMiddleware))
        fsExtra.ensureDirSync(this.settings.upload.temp)
    }
    private createUploadHandler(middleware: multer.Instance) {
        let receive = middleware.single("file")
        return (req, res) => {
            receive(req, res, (err) => {
                if (err) {
                    res.status(400)
                    this.error(res, new Errors.UploadError("Lib error"))
                    return
                }
                if (!req.file) {
                    res.status(400)
                    this.error(res, new Errors.UploadError("No file"))
                    return
                }
                let file = req.file as Express.Multer.File
                let path = file.path
                let originalName = file.originalname
                fs.stat(path, (err, stat) => {
                    if (err) {
                        res.status(500)
                        this.error(res, new Errors.UploadError("Fail to find temp file", { via: err }))
                        return
                    }
                    this.services.BlobService.createBlob({ path, name: originalName }, (err, blob) => {
                        if (err) {
                            res.status(500)
                            this.error(res, new Errors.UploadError("Fail to create blob", { via: err }))
                            return
                        }
                        this.success(res, blob)
                    })

                })
            })
        }
    }
    private createMulterConfig() {
        let storage = multer.diskStorage({
            destination: (req, file, cb) => {
                cb(null, this.settings.upload.temp)
            },
            filename: (req, file, cb) => {
                let fileFormat = (file.originalname).split('.');
                let ext = pathModule.extname(file.originalname)
                let whitelist = this.settings.upload.acceptedExtnames
                if (whitelist && whitelist.length > 0) {
                    if (whitelist.indexOf(ext) < 0) {
                        cb(new Error("Extname not supported"), null)
                        return
                    }
                }
                if (!ext || ext.length > 8) {
                    ext = ".unknown"
                }
                let rand = Date.now() + Math.random().toString().slice(4, 12)
                cb(null, rand + ext)
            },
        })
        return storage
    }
}
